# FITS-Demo.py
# Creates, writes, reads, displays, and fits a simple rectangular image FITS file.
# M.Lampton UCB SSL July2016  refreshed August 2018

from astropy.io import fits
import numpy as np
import matplotlib.pyplot as plt  
import matplotlib.patches as patches
from scipy.optimize import curve_fit

#----------declare constants------------------------
#  These appear only on RHSides and are therefore automatically global.
#  I use all UPPERCASE for these global constants.

NCOLS = 100   # x axis
NROWS = 128   # y axis
NPIXELS = NCOLS*NROWS
NPARMS = 5

#-------declare functions----------------------------

def gaussPSF(r, fwhm):
# Here I model our PSF as a 2D Gaussian with given FWHM in pixels.
# "r" is the radial distance from the center of this Gaussian, in pixels.
# Normalized to make peak=1
    rms = fwhm/2.35
    if rms <= 0:
      return 0.
    arg = 0.5*(r/rms)**2
    if arg > 5:
      return 0.
    z = np.exp(-arg)
    return z

def myFunc2D(parms):
    # Produces a noiseless image, NROWS x NCOLS, from given parameters.
    # parms[]: {background/pix;  peakSignal/pix; xpeak; ypeak; fwhmpix}
    result = np.zeros((NROWS,NCOLS))
    for row in range(NROWS):
        for col in range(NCOLS):
            dx = col - parms[2]
            dy = row - parms[3]
            r = np.sqrt(dx**2 + dy**2) 
            result[row,col] = parms[0] + parms[1]*gaussPSF(r, parms[4])
    return result
        
#---------oblique plot package starts here----------------

SHELF  = 5.0          # pedestal height above zero, screen coords
FLOOR  = -5.0         # floor of polygons, screen coords
HEIGHT = 70.0         # scale image peak to screen coords
     
def getHoriz(x,y):    # horizontal screen coordinate
    return x+y
    
def getVert(x,y,z):   # vertical screen coordinate
    return -0.4*x + 0.4*y + z

def showOblique():    # scales and plots thumb[i,j]
    zmax = np.amax(thumb)
    zmin = np.amin(thumb)
    span = zmax-zmin
    coef = HEIGHT/span
    fig = plt.figure(figsize=(6,5))
    ax = fig.add_subplot(111)
    toplim = 0
    botlim = 0
    leftlim = 0
    rightlim = 0
    for row in range(0, NROWS):    # one 2D jagged "trapezoid" per row
        trapezoid = np.zeros(shape=(NCOLS+2, 2))
        for col in range(0, NCOLS):
            z = coef*(thumb[row,col] - zmin) 
            trapezoid[col] = (getHoriz(row,col), getVert(row,col,z))
        trapezoid[NCOLS] = (getHoriz(row, NCOLS-1), getVert(row, NCOLS-1,FLOOR))
        trapezoid[NCOLS+1] = (getHoriz(row,0), getVert(row,0,FLOOR)) 
        rightlim = max(rightlim, int(np.amax(trapezoid[:,0])))
        leftlim = min(leftlim, int(np.amin(trapezoid[:,0])))
        toplim = max(toplim, int(np.amax(trapezoid[:,1])))
        botlim = min(botlim, int(np.amin(trapezoid[:,1])))
        poly = patches.Polygon(trapezoid, facecolor=(0.,0.,1.), edgecolor=(1.,1.,1))
        ax.add_patch(poly)
    ax.set_xlim(leftlim-10, rightlim+10)
    ax.set_ylim(botlim-10, toplim+10)
    plt.savefig('FITS-Demo-Oblique.png')
    plt.show()
    
#---Optimization helpers: user should call minimize() not curve_fit()-----
    
def minimize2D(func2D, goals2D, pinit1D, sigmas2D):
    # Converts user 2D arrays into 1D for calling curve_fit().
    # Also provides required args1D[] but we ignore them and use pixel {row,col}
    args1D = np.arange(NPIXELS) 
    goals1D = goals2D.reshape(NPIXELS)
    sigmas1D = sigmas2D.reshape(NPIXELS)
    popt, pcov = curve_fit(wrapfunc1D, args1D, goals1D, pinit1D, sigmas1D)
    return popt, pcov
    
def wrapfunc1D(args, *parms): 
    # Removing the asterisk below gives us our array.
    # args[] is discarded since we use pixel {row,col} as arguments.  
    global ncalls
    ncalls += 1
    return np.reshape(myFunc2D(parms), NPIXELS)


#--------------------------main program---------------------
#--------------------------main program---------------------

#----First, define a star image atop a flat background------

#-------------------bkg----peak----xave----yave----fwhm------
myPtrue = np.array([5.0,   20.0,   60.0,   70.0,   8.0])
myPlabels =      ['back', 'peak', 'xave', 'yave', 'fwhm']

#----Second, add noise and save the image as a .FITS file---

filename = 'FITS-Demo.fits'
myHDU = fits.PrimaryHDU()
myHDU.data = myFunc2D(myPtrue)
NOISE = 1.0
myHDU.data += np.random.normal(0., NOISE, (NROWS,NCOLS))
myHDU.writeto(filename, overwrite=True)


#-------Third: read that noisy file back in------------------

hdulist = fits.open(filename)
print(hdulist.info())
hdu = hdulist[0]
thumb = hdu.data
print('Rows, cols:   ', thumb.shape)
print('Sum ADU:     {:12.3f}'.format(np.sum(thumb)))
print('Average ADU: {:12.3f}'.format(np.mean(thumb)))
print('Median ADU:  {:12.3f}'.format(np.median(thumb)))
print('Max ADU:     {:12.3f}'.format(np.max(thumb)))
print('Min ADU:     {:12.3f}'.format(np.min(thumb)))
#...and get the peak location...
peakrow, peakcol = np.unravel_index(thumb.argmax(), thumb.shape)
print('peak col = xpeak:  {:6d}'.format(peakcol))
print('peak row = ypeak:  {:6d}'.format(peakrow))
print('peak value   {:12.3f}'.format(thumb[peakrow,peakcol]))


#-------Fourth: plot a color-coded noisy image---------------

fig = plt.figure(figsize=(6,6))   # inches
image = plt.imshow(thumb)
plt.title(filename)
plt.savefig('FITS-Demo-Image.png')
plt.show()


#----Fifth: show an oblique view of the noisy image------------

showOblique()


#-------Let's fit a Gaussian to the noisy image----------------

print('\nStarting fit to the noisy Gaussian...')
myPinit = np.array([10.,10., peakcol, peakrow, 10.]) # an initial guess
mySigmas2D = NOISE*np.ones([NROWS,NCOLS])            # array of expected errors
myData2D = thumb                                     # the noisy image 
ncalls = 0
optim, covar = minimize2D(myFunc2D, myData2D, myPinit, mySigmas2D)
print('ncalls = ', ncalls)
print('TrueParms:   {:10.4f} {:10.4f} {:10.4f} {:10.4f} {:10.4f}'.format(*myPtrue))
print('BestParms:   {:10.4f} {:10.4f} {:10.4f} {:10.4f} {:10.4f}'.format(*optim))


#-----Finally let's gather many noisy images and get their statistics-------

trueImage2D = myFunc2D(myPtrue)
NITERS = 10
NOISE = 1.0
parmstats = np.zeros([NITERS, NPARMS])
print('\nStarting the iterative fitting run, NITERS = ', NITERS)

print('--iter-ncalls--background---peak--------xbest--------ybest--------fwhm-----')
for iter in range(NITERS):
    ncalls = 0
    noiseImage2D = np.random.normal(0, NOISE, (NROWS,NCOLS))
    noisyImage2D = trueImage2D + noiseImage2D
    popt, covar = minimize2D(myFunc2D, noisyImage2D, myPinit, mySigmas2D)
    print('{:6d}{:6d}'.format(iter, ncalls), end='')
    for parm in popt:
        print('{:12.6f}'.format(parm), end='')
    print()
    for jparm in range(5):
        parmstats[iter,jparm] = popt[jparm]
print('\nMeans and RMSs for the parameters:')
for jparm in range(NPARMS):
    mean = np.average(parmstats[:,jparm])
    rms  = np.std(parmstats[:,jparm])
    label = myPlabels[jparm]
    print('parameter, mean, rms = {:6s}{:12.6f}{:12.6f}'.format(label, mean, rms))
    